@@ -31,9 +31,12 @@
 #include "defs.h"
 #include "host_defs.h"
 #include "h_extern.h"
+#include "attributes.h"
+#include "p11util.h"
 #include "tok_spec_struct.h"
 #include "pkcs32.h"
 #include "trace.h"
+#include "../api/policy.h"
 
 // object_create()
 //
@@ -51,12 +54,9 @@ CK_RV object_create(STDLL_TokData_t * tokdata,
                     CK_ATTRIBUTE * pTemplate, CK_ULONG ulCount, OBJECT ** obj)
 {
     OBJECT *o = NULL;
-    CK_ATTRIBUTE *attr = NULL;
-    CK_BBOOL class_given = FALSE;
     CK_BBOOL subclass_given = FALSE;
-    CK_ULONG class = 0xFFFFFFFF, subclass = 0xFFFFFFFF;
+    CK_ULONG class, subclass = 0xFFFFFFFF;
     CK_RV rc;
-    unsigned int i;
 
     if (!pTemplate) {
         TRACE_ERROR("Invalid function arguments.\n");
@@ -64,33 +64,39 @@ CK_RV object_create(STDLL_TokData_t * tokdata,
     }
     // extract the object class and subclass
     //
-    attr = pTemplate;
-    for (i = 0; i < ulCount; i++, attr++) {
-        if (attr->type == CKA_CLASS) {
-            class = *(CK_OBJECT_CLASS *) attr->pValue;
-            class_given = TRUE;
-        }
-
-        if (attr->type == CKA_CERTIFICATE_TYPE) {
-            subclass = *(CK_CERTIFICATE_TYPE *) attr->pValue;
-            subclass_given = TRUE;
-        }
-
-        if (attr->type == CKA_KEY_TYPE) {
-            subclass = *(CK_KEY_TYPE *) attr->pValue;
-            subclass_given = TRUE;
-        }
-
-        if (attr->type == CKA_HW_FEATURE_TYPE) {
-            subclass = *(CK_HW_FEATURE_TYPE *) attr->pValue;
-            subclass_given = TRUE;
-        }
+    rc = get_ulong_attribute_by_type(pTemplate, ulCount, CKA_CLASS,
+                                     &class);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("Could not find CKA_CLASS for the key.\n");
+        return rc;
     }
 
-    if (class_given == FALSE) {
-        TRACE_ERROR("%s\n", ock_err(ERR_TEMPLATE_INCOMPLETE));
-        return CKR_TEMPLATE_INCOMPLETE;
+    rc = get_ulong_attribute_by_type(pTemplate, ulCount, CKA_CERTIFICATE_TYPE,
+                                     &subclass);
+    if (rc == CKR_ATTRIBUTE_VALUE_INVALID) {
+        TRACE_ERROR("%s\n", ock_err(ERR_ATTRIBUTE_VALUE_INVALID));
+        return CKR_ATTRIBUTE_VALUE_INVALID;
     }
+    if (rc == CKR_OK)
+        subclass_given = TRUE;
+
+    rc = get_ulong_attribute_by_type(pTemplate, ulCount, CKA_KEY_TYPE,
+                                     &subclass);
+    if (rc == CKR_ATTRIBUTE_VALUE_INVALID) {
+        TRACE_ERROR("%s\n", ock_err(ERR_ATTRIBUTE_VALUE_INVALID));
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+    }
+    if (rc == CKR_OK)
+        subclass_given = TRUE;
+
+    rc = get_ulong_attribute_by_type(pTemplate, ulCount, CKA_HW_FEATURE_TYPE,
+                                     &subclass);
+    if (rc == CKR_ATTRIBUTE_VALUE_INVALID) {
+        TRACE_ERROR("%s\n", ock_err(ERR_ATTRIBUTE_VALUE_INVALID));
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+    }
+    if (rc == CKR_OK)
+        subclass_given = TRUE;
 
     // Return CKR_ATTRIBUTE_TYPE_INVALID when trying to create a
     // vendor-defined object.
@@ -99,9 +105,10 @@ CK_RV object_create(STDLL_TokData_t * tokdata,
         return CKR_ATTRIBUTE_TYPE_INVALID;
     }
 
-    if (class != CKO_DATA && subclass_given != TRUE) {
-        TRACE_ERROR("%s\n", ock_err(ERR_TEMPLATE_INCOMPLETE));
-        return CKR_TEMPLATE_INCOMPLETE;
+    if (subclass_given != TRUE
+        && class != CKO_DATA && class != CKO_PROFILE) {
+        TRACE_ERROR("%s\n", ock_err(ERR_TEMPLATE_INCONSISTENT));
+        return CKR_TEMPLATE_INCONSISTENT;
     }
 
     rc = object_create_skel(tokdata, pTemplate, ulCount,
@@ -131,12 +138,13 @@ CK_RV object_create(STDLL_TokData_t * tokdata,
 // the individual object level don't have the concept of "session". These checks
 // are done by the object manager.
 //
+// The old_obj must hold the READ lock!
+//
 CK_RV object_copy(STDLL_TokData_t * tokdata,
                   CK_ATTRIBUTE * pTemplate,
                   CK_ULONG ulCount, OBJECT * old_obj, OBJECT ** new_obj)
 {
-    TEMPLATE *tmpl = NULL;
-    TEMPLATE *new_tmpl = NULL;
+    TEMPLATE *tmpl, *new_tmpl;
     OBJECT *o = NULL;
     CK_BBOOL found;
     CK_ULONG class, subclass;
@@ -167,14 +175,20 @@ CK_RV object_copy(STDLL_TokData_t * tokdata,
     memset(o, 0x0, sizeof(OBJECT));
     memset(tmpl, 0x0, sizeof(TEMPLATE));
     memset(new_tmpl, 0x0, sizeof(TEMPLATE));
+    o->template = tmpl;
+
+    rc = object_init_lock(o);
+    if (rc != CKR_OK)
+        goto error;
 
     // copy the original object's attribute template
     //
-    rc = template_copy(tmpl, old_obj->template);
+    rc = template_copy(o->template, old_obj->template);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Failed to copy template.\n");
         goto error;
     }
+
     rc = template_add_attributes(new_tmpl, pTemplate, ulCount);
     if (rc != CKR_OK) {
         TRACE_DEVEL("template_add_attributes failed.\n");
@@ -188,7 +202,7 @@ CK_RV object_copy(STDLL_TokData_t * tokdata,
     //    4) conflicting attributes/values
     //
 
-    found = template_get_class(tmpl, &class, &subclass);
+    found = template_get_class(o->template, &class, &subclass);
     if (found == FALSE) {
         TRACE_ERROR("Could not find CKA_CLASS in object's template.\n");
         rc = CKR_TEMPLATE_INCONSISTENT;
@@ -208,9 +222,23 @@ CK_RV object_copy(STDLL_TokData_t * tokdata,
         TRACE_DEVEL("template_validate_attributes failed.\n");
         goto error;
     }
+
+    /*
+     * Let the token know that attributes of the copied object have been
+     * changed.
+     */
+    if (token_specific.t_set_attribute_values != NULL) {
+        rc = token_specific.t_set_attribute_values(tokdata, o, new_tmpl);
+        if (rc != CKR_OK) {
+            TRACE_DEVEL("token_specific_set_attribute_values failed with %lu\n",
+                        rc);
+            goto error;
+        }
+   }
+
     // merge in the new attributes
     //
-    rc = template_merge(tmpl, &new_tmpl);
+    rc = template_merge(o->template, &new_tmpl);
     if (rc != CKR_OK) {
         TRACE_DEVEL("template_merge failed.\n");
         goto error;
@@ -219,21 +247,19 @@ CK_RV object_copy(STDLL_TokData_t * tokdata,
     // object's template (contained in tmpl) already has the required attributes
     // present
     //
-    rc = template_check_required_attributes(tmpl, class, subclass, MODE_COPY);
+    rc = template_check_required_attributes(o->template, class, subclass,
+                                            MODE_COPY);
     if (rc != CKR_OK) {
         TRACE_ERROR("template_check_required_attributes failed.\n");
         goto error;
     }
     // at this point, we should have a valid object with correct attributes
     //
-    o->template = tmpl;
     *new_obj = o;
 
     return CKR_OK;
 
 error:
-    if (tmpl)
-        template_free(tmpl);
     if (new_tmpl)
         template_free(new_tmpl);
     if (o)
@@ -306,41 +332,57 @@ void object_free(OBJECT * obj)
     if (obj) {
         if (obj->template)
             template_free(obj->template);
+        object_destroy_lock(obj);
         free(obj);
     }
 }
 
-//call_free()
+//call_object_free()
 //This function is added to silence the compiler during implicit void (*)(void*)
 //function pointer casting in call back functions.
 //
-void call_free(void *ptr)
+void call_object_free(void *ptr)
 {
     if (ptr)
         object_free((OBJECT *) ptr);
 }
 
-// object_is_modifiable()
-//
 CK_BBOOL object_is_modifiable(OBJECT * obj)
 {
-    CK_ATTRIBUTE *attr = NULL;
-    CK_BBOOL modifiable;
-    CK_BBOOL found;
+    CK_BBOOL val;
+    CK_RV rc;
 
-    found = template_attribute_find(obj->template, CKA_MODIFIABLE, &attr);
-    if (found == FALSE)
-        return TRUE;            // should always be found but we default to TRUE
+    rc = template_attribute_get_bool(obj->template, CKA_MODIFIABLE, &val);
+    if (rc != CKR_OK)
+        return TRUE;
 
-    //axelrh: prevent dereferencing NULL from bad parse
-    if (attr->pValue == NULL)
-        return TRUE;            //default to TRUE
+    return val;
 
-    modifiable = *(CK_BBOOL *) attr->pValue;
-
-    return modifiable;
 }
 
+CK_BBOOL object_is_copyable(OBJECT * obj)
+{
+    CK_BBOOL val;
+    CK_RV rc;
+
+    rc = template_attribute_get_bool(obj->template, CKA_COPYABLE, &val);
+     if (rc != CKR_OK)
+         return TRUE;
+
+    return val;
+}
+
+CK_BBOOL object_is_destroyable(OBJECT * obj)
+{
+    CK_BBOOL val;
+    CK_RV rc;
+
+    rc = template_attribute_get_bool(obj->template, CKA_DESTROYABLE, &val);
+     if (rc != CKR_OK)
+         return TRUE;
+
+    return val;
+}
 
 // object_is_private()
 //
@@ -348,27 +390,14 @@ CK_BBOOL object_is_modifiable(OBJECT * obj)
 //
 CK_BBOOL object_is_private(OBJECT * obj)
 {
-    CK_ATTRIBUTE *attr = NULL;
-    CK_BBOOL priv;
-    CK_BBOOL found;
+    CK_BBOOL val;
+    CK_RV rc;
 
-    found = template_attribute_find(obj->template, CKA_PRIVATE, &attr);
-    if (found == FALSE) {
-        return TRUE;            // should always be found but we default to TRUE
-    }
-    if (attr == NULL)
-        return TRUE;
+    rc = template_attribute_get_bool(obj->template, CKA_PRIVATE, &val);
+     if (rc != CKR_OK)
+         return TRUE;
 
-
-    //axelrh: prevent segfault caused by 0-len attribute
-    //that has a null pValue
-    CK_BBOOL *bboolPtr = (CK_BBOOL *) attr->pValue;
-    if (bboolPtr == NULL)
-        return TRUE;            //default
-
-    priv = *(bboolPtr);
-
-    return priv;
+    return val;
 }
 
 
@@ -391,21 +420,14 @@ CK_BBOOL object_is_public(OBJECT * obj)
 //
 CK_BBOOL object_is_token_object(OBJECT * obj)
 {
-    CK_ATTRIBUTE *attr = NULL;
-    CK_BBOOL is_token;
-    CK_BBOOL found;
+    CK_BBOOL val;
+    CK_RV rc;
 
-    found = template_attribute_find(obj->template, CKA_TOKEN, &attr);
-    if (found == FALSE)
+    rc = template_attribute_get_bool(obj->template, CKA_TOKEN, &val);
+    if (rc != CKR_OK)
         return FALSE;
 
-    //axelrh: prevent dereferencing NULL from bad parse
-    if (attr->pValue == NULL)
-        return FALSE;
-
-    is_token = *(CK_BBOOL *) attr->pValue;
-
-    return is_token;
+    return val;
 }
 
 
@@ -423,6 +445,47 @@ CK_BBOOL object_is_session_object(OBJECT * obj)
         return TRUE;
 }
 
+// object_is_extractable()
+//
+CK_BBOOL object_is_extractable(OBJECT * obj)
+{
+    CK_BBOOL val;
+    CK_RV rc;
+
+    rc = template_attribute_get_bool(obj->template, CKA_EXTRACTABLE, &val);
+    if (rc != CKR_OK)
+        return TRUE;
+
+    return val;
+}
+
+// object_is_pkey_extractable()
+//
+CK_BBOOL object_is_pkey_extractable(OBJECT * obj)
+{
+    CK_BBOOL val;
+    CK_RV rc;
+
+    rc = template_attribute_get_bool(obj->template, CKA_IBM_PROTKEY_EXTRACTABLE, &val);
+    if (rc != CKR_OK)
+        return FALSE;
+
+    return val;
+}
+
+// object_is_attr_bound()
+//
+CK_BBOOL object_is_attr_bound(OBJECT * obj)
+{
+    CK_BBOOL val;
+    CK_RV rc;
+
+    rc = template_attribute_get_bool(obj->template, CKA_IBM_ATTRBOUND, &val);
+    if (rc != CKR_OK)
+        return FALSE;
+
+    return val;
+}
 
 // object_get_size()
 //
@@ -435,6 +498,63 @@ CK_ULONG object_get_size(OBJECT * obj)
     return size;
 }
 
+static CK_RV object_get_attribute_array(CK_ATTRIBUTE *array_attr,
+                                        CK_ATTRIBUTE *tmpl_attr)
+{
+    CK_RV rc = CKR_OK, rc2;
+    CK_ULONG num_elemets, i;
+    CK_ATTRIBUTE_PTR array_elements;
+    CK_ATTRIBUTE_PTR tmpl_elements;
+
+    if (!is_attribute_attr_array(array_attr->type))
+        return CKR_ATTRIBUTE_TYPE_INVALID;
+
+    if (tmpl_attr->pValue == NULL) {
+        tmpl_attr->ulValueLen = array_attr->ulValueLen;
+        return CKR_OK;
+    }
+    if (tmpl_attr->ulValueLen < array_attr->ulValueLen) {
+        tmpl_attr->ulValueLen = CK_UNAVAILABLE_INFORMATION;
+        return CKR_BUFFER_TOO_SMALL;
+    }
+
+    num_elemets = array_attr->ulValueLen / sizeof(CK_ATTRIBUTE);
+    array_elements = (CK_ATTRIBUTE_PTR)array_attr->pValue;
+    tmpl_elements = (CK_ATTRIBUTE_PTR)tmpl_attr->pValue;
+
+    for (i = 0; i < num_elemets; i++) {
+        tmpl_elements[i].type = array_elements[i].type;
+
+        if (tmpl_elements[i].pValue == NULL) {
+            tmpl_elements[i].ulValueLen = array_elements[i].ulValueLen;
+        } else if (tmpl_elements[i].ulValueLen >=
+                                            array_elements[i].ulValueLen) {
+            if (array_elements[i].pValue != NULL) {
+                if (is_attribute_attr_array(array_elements[i].type)) {
+                    rc2 = object_get_attribute_array(&array_elements[i],
+                                                     &tmpl_elements[i]);
+                    if (rc2 == CKR_BUFFER_TOO_SMALL)
+                        rc = rc2;
+                    else if (rc2 != CKR_OK) {
+                        TRACE_ERROR("object_get_attribute_array failed\n");
+                        rc = rc2;
+                        break;
+                    }
+                } else {
+                    memcpy(tmpl_elements[i].pValue, array_elements[i].pValue,
+                           array_elements[i].ulValueLen);
+                }
+            }
+            tmpl_elements[i].ulValueLen = array_elements[i].ulValueLen;
+        } else {
+            TRACE_ERROR("%s\n", ock_err(ERR_BUFFER_TOO_SMALL));
+            rc = CKR_BUFFER_TOO_SMALL;
+            tmpl_elements[i].ulValueLen = CK_UNAVAILABLE_INFORMATION;
+        }
+    }
+
+    return rc;
+}
 
 //
 //
@@ -445,7 +565,7 @@ CK_RV object_get_attribute_values(OBJECT * obj,
     CK_ATTRIBUTE *attr = NULL;
     CK_ULONG i;
     CK_BBOOL flag;
-    CK_RV rc;
+    CK_RV rc, rc2;
 
     rc = CKR_OK;
 
@@ -454,29 +574,44 @@ CK_RV object_get_attribute_values(OBJECT * obj,
     for (i = 0; i < ulCount; i++) {
         flag = template_check_exportability(obj_tmpl, pTemplate[i].type);
         if (flag == FALSE) {
-            TRACE_ERROR("%s\n", ock_err(ERR_ATTRIBUTE_SENSITIVE));
+            TRACE_ERROR("%s: %lx\n", ock_err(ERR_ATTRIBUTE_SENSITIVE),
+                        pTemplate[i].type);
             rc = CKR_ATTRIBUTE_SENSITIVE;
-            pTemplate[i].ulValueLen = (CK_ULONG) - 1;
+            pTemplate[i].ulValueLen = (CK_ULONG)CK_UNAVAILABLE_INFORMATION;
             continue;
         }
 
         flag = template_attribute_find(obj_tmpl, pTemplate[i].type, &attr);
         if (flag == FALSE) {
-            TRACE_ERROR("%s\n", ock_err(ERR_ATTRIBUTE_TYPE_INVALID));
+            TRACE_ERROR("%s: %lx\n", ock_err(ERR_ATTRIBUTE_TYPE_INVALID),
+                        pTemplate[i].type);
             rc = CKR_ATTRIBUTE_TYPE_INVALID;
-            pTemplate[i].ulValueLen = (CK_ULONG) - 1;
+            pTemplate[i].ulValueLen = (CK_ULONG)CK_UNAVAILABLE_INFORMATION;
             continue;
         }
 
         if (pTemplate[i].pValue == NULL) {
             pTemplate[i].ulValueLen = attr->ulValueLen;
         } else if (pTemplate[i].ulValueLen >= attr->ulValueLen) {
-            memcpy(pTemplate[i].pValue, attr->pValue, attr->ulValueLen);
+            if (attr->pValue != NULL) {
+                if (is_attribute_attr_array(attr->type)) {
+                    rc2 = object_get_attribute_array(attr, &pTemplate[i]);
+                    if (rc2 == CKR_BUFFER_TOO_SMALL)
+                        rc = rc2;
+                    else if (rc2 != CKR_OK) {
+                        TRACE_ERROR("object_get_attribute_array failed\n");
+                        rc = rc2;
+                        break;
+                    }
+                } else {
+                    memcpy(pTemplate[i].pValue, attr->pValue, attr->ulValueLen);
+                }
+            }
             pTemplate[i].ulValueLen = attr->ulValueLen;
         } else {
             TRACE_ERROR("%s\n", ock_err(ERR_BUFFER_TOO_SMALL));
             rc = CKR_BUFFER_TOO_SMALL;
-            pTemplate[i].ulValueLen = (CK_ULONG) - 1;
+            pTemplate[i].ulValueLen = (CK_ULONG)CK_UNAVAILABLE_INFORMATION;
         }
     }
 
@@ -532,6 +667,16 @@ CK_RV object_set_attribute_values(STDLL_TokData_t * tokdata,
         TRACE_DEVEL("template_validate_attributes failed.\n");
         goto error;
     }
+
+    if (token_specific.t_set_attribute_values != NULL) {
+        rc = token_specific.t_set_attribute_values(tokdata, obj, new_tmpl);
+        if (rc != CKR_OK) {
+            TRACE_DEVEL("token_specific_set_attribute_values failed with %lu\n",
+                        rc);
+            goto error;
+        }
+   }
+
     // merge in the new attributes
     //
     rc = template_merge(obj->template, &new_tmpl);
@@ -553,18 +698,13 @@ error:
 }
 
 
-//
-//
-CK_RV object_restore(CK_BYTE * data, OBJECT ** new_obj, CK_BBOOL replace)
-{
-    return object_restore_withSize(data, new_obj, replace, -1);
-}
-
 //
 //Modified object_restore to prevent buffer overflow
 //If data_size=-1, won't do bounds checking
-CK_RV object_restore_withSize(CK_BYTE * data, OBJECT ** new_obj,
-                              CK_BBOOL replace, int data_size)
+CK_RV object_restore_withSize(struct policy *policy,
+                              CK_BYTE * data, OBJECT ** new_obj,
+                              CK_BBOOL replace, int data_size,
+                              const char *fname)
 {
     TEMPLATE *tmpl = NULL;
     OBJECT *obj = NULL;
@@ -572,6 +712,7 @@ CK_RV object_restore_withSize(CK_BYTE * data, OBJECT ** new_obj,
     CK_ULONG_32 count = 0;
     CK_RV rc;
     CK_OBJECT_CLASS_32 class32;
+    const char *obj_name;
 
     if (!data || !new_obj) {
         TRACE_ERROR("Invalid function arguments.\n");
@@ -584,7 +725,6 @@ CK_RV object_restore_withSize(CK_BYTE * data, OBJECT ** new_obj,
         goto error;
     }
 
-
     memset(obj, 0x0, sizeof(OBJECT));
 
     memcpy(&class32, data + offset, sizeof(CK_OBJECT_CLASS_32));
@@ -594,23 +734,65 @@ CK_RV object_restore_withSize(CK_BYTE * data, OBJECT ** new_obj,
     memcpy(&count, data + offset, sizeof(CK_ULONG_32));
     offset += sizeof(CK_ULONG_32);
 
-
     memcpy(&obj->name, data + offset, 8);
     offset += 8;
 
+    if (fname != NULL) {
+        /* The last path element of the file name must match the object name */
+        obj_name = strrchr(fname, '/');
+        if (obj_name == NULL) {
+            TRACE_ERROR("File name has invalid format: '%s'\n", fname);
+            rc = CKR_FUNCTION_FAILED;
+            goto error;
+        }
+
+        obj_name++;
+        if (strlen(obj_name) != 8) {
+            TRACE_ERROR("File name has invalid format: '%s'\n", fname);
+            rc = CKR_FUNCTION_FAILED;
+            goto error;
+        }
+
+        if (memcmp(obj->name, obj_name, 8) != 0) {
+            TRACE_ERROR("Object name '%.8s' does not match the file name it was loaded from: '%s'\n",
+                        obj->name, fname);
+            rc = CKR_FUNCTION_FAILED;
+            goto error;
+        }
+    }
+
     rc = template_unflatten_withSize(&tmpl, data + offset, count, data_size);
     if (rc != CKR_OK) {
         TRACE_DEVEL("template_unflatten_withSize failed.\n");
         goto error;
     }
+    /* External tools (e.g., pkcscca) might use this function and not
+       be aware of any policy.  Allow them to pass NULL. */
+    if (policy) {
+        /* Ignore policy violations here since the point is to get the
+           correct strength classification for the usage scenario
+           which will then allow or block key usage. */
+        policy->store_object_strength(policy, &obj->strength,
+                                      policy_get_attr_from_template,
+                                      tmpl, NULL, NULL);
+    }
+
     obj->template = tmpl;
+    tmpl = NULL;
 
     if (replace == FALSE) {
+        rc = object_init_lock(obj);
+        if (rc != CKR_OK)
+            goto error;
+
         *new_obj = obj;
     } else {
+        /* Reload of existing object only changes the template */
         template_free((*new_obj)->template);
-        memcpy(*new_obj, obj, sizeof(OBJECT));
-
+        (*new_obj)->template = obj->template;
+        (*new_obj)->strength.strength = obj->strength.strength;
+        (*new_obj)->strength.siglen = obj->strength.siglen;
+        (*new_obj)->strength.allowed = obj->strength.allowed;
         free(obj);              // don't want to do object_free() here!
     }
 
@@ -687,6 +869,14 @@ CK_RV object_create_skel(STDLL_TokData_t * tokdata,
     if (rc != CKR_OK)
         goto done;
 
+    if (token_specific.t_set_attrs_for_new_object != NULL) {
+        rc = token_specific.t_set_attrs_for_new_object(tokdata, class,
+                                                       mode, tmpl2);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("token_specific.t_set_pkey_attr failed with rc=%lx\n",rc);
+            goto done;
+        }
+    }
 
     rc = template_merge(tmpl, &tmpl2);
     if (rc != CKR_OK) {
@@ -696,6 +886,12 @@ CK_RV object_create_skel(STDLL_TokData_t * tokdata,
     // at this point, we should have a valid object with correct attributes
     //
     o->template = tmpl;
+    tmpl = NULL;
+
+    rc = object_init_lock(o);
+    if (rc != CKR_OK)
+        goto done;
+
     *obj = o;
 
     return CKR_OK;
@@ -710,3 +906,60 @@ done:
 
     return rc;
 }
+
+CK_RV object_init_lock(OBJECT *obj)
+{
+    if (pthread_rwlock_init(&obj->template_rwlock, NULL) != 0) {
+        TRACE_DEVEL("Object Lock init failed.\n");
+        return CKR_CANT_LOCK;
+    }
+
+    return CKR_OK;
+}
+
+CK_RV object_destroy_lock(OBJECT *obj)
+{
+    if (pthread_rwlock_destroy(&obj->template_rwlock) != 0) {
+        TRACE_DEVEL("Object Lock destroy failed.\n");
+        return CKR_CANT_LOCK;
+    }
+
+    return CKR_OK;
+}
+
+/*
+ * Do NOT try to get an object lock, if the current thread holds the
+ * XProcLock! This might case a deadlock !
+ * Always first acquire the Object lock, and then the XProcLock.
+ */
+CK_RV object_lock(OBJECT *obj, OBJ_LOCK_TYPE type)
+{
+    switch (type) {
+    case NO_LOCK:
+        break;
+    case READ_LOCK:
+        if (pthread_rwlock_rdlock(&obj->template_rwlock) != 0) {
+            TRACE_DEVEL("Object Read-Lock failed.\n");
+            return CKR_CANT_LOCK;
+        }
+        break;
+    case WRITE_LOCK:
+        if (pthread_rwlock_wrlock(&obj->template_rwlock) != 0) {
+            TRACE_DEVEL("Object Write-Lock failed.\n");
+            return CKR_CANT_LOCK;
+        }
+        break;
+    }
+
+    return CKR_OK;
+}
+
+CK_RV object_unlock(OBJECT *obj)
+{
+    if (pthread_rwlock_unlock(&obj->template_rwlock) != 0) {
+        TRACE_DEVEL("Object Unlock failed.\n");
+        return CKR_CANT_LOCK;
+    }
+
+    return CKR_OK;
+}
